from freezedata import freeze_data
import types
from collections import namedtuple

import pytest

def test_freeze_data_lists_sets_dicts():
    data = [1, [1, {1, 2, 3}], {1: 2, 2: [1, 2]},
            types.MappingProxyType({1: 2, 2: (1, 2)})]
    control_data = (1, (1, frozenset((1, 2, 3))), types.MappingProxyType({1: 2, 2: (1, 2)}),
                    types.MappingProxyType({1: 2, 2: (1, 2)}))

    assert freeze_data(data) == control_data


def test_freeze_data_namedtuples_simplenamespaces():
    TestNamedTuple = namedtuple('TestNamedTuple', list('abc'))
    data = types.SimpleNamespace(a=1, b=TestNamedTuple(a=1, b=2, c=3))

    control_tuple = namedtuple('SimpleNamespace', list('ab'))
    control_data = control_tuple(a=1, b=TestNamedTuple(a=1, b=2, c=3))

    assert freeze_data(data) == control_data


def test_freeze_data_functions():

    def test_func(n):
        return n*2
    data = [test_func]
    frozendata = freeze_data(data, allow='functions')

    with pytest.raises(ValueError):
        frozendata2 = freeze_data(data)

    from freezedata.freezedata import copy_function
    control_data = (copy_function(test_func),)

    assert frozendata[0](2) == test_func(2)

    assert type(frozendata) == tuple

    assert type(frozendata[0]) == types.FunctionType

    assert frozendata[0] != control_data[0]


def test_freeze_data_class_instance():
    from . import data_module

    with pytest.raises(ValueError):
        frozendata2 = freeze_data(data_module)
    frozendata = freeze_data(data_module, allow={'classes', 'instances'})


"""
Recursively convert ``list`` to ``tuple``, ``set`` to ``frozenset``,
``dict`` to ``mappingproxy`` etc.

Example usage:

.. code-block:: python

    import freezedata

    data = [{'a': [1,2,3], 'b': {1,2,3}}, {1:1, 2:2, 3:3}]
    frozendata = freezedata.freeze_data(data)
    print(frozendata)
    >> (mappingproxy({'a': (1, 2, 3), 'b': frozenset({1, 2, 3})}), mappingproxy({1: 1, 2: 2, 3: 3}))

This is a read-only data structure, that is; there is no direct way to alter this
data structure from within ``frozendata`` itself (without using some special modules (``gc``,
``inspect``)).

For example:

.. code-block:: python

    frozendata[0]['a'][0] = 4
    >> TypeError: 'tuple' object does not support item assignment
    del frozendata[1][1]
    >> TypeError: 'mappingproxy' object does not support item deletion

*Notice*: Since a ``mappingproxy`` is not hashable, frozen data
structures containing ``mappingproxy`` (i.e. based on ``dict``) will not be
hashable either:

.. code-block:: python

    hash(frozendata)
    >> TypeError: unhashable type: 'mappingproxy'

On the other hand, if the frozen data structure contains only hashable elements, the whole
structure will be hashable (and immutable) as well:

.. code-block:: python

    frozendata = freezedata.freeze_data([[1,2,3], {4,5,6}])
    print(frozendata)
    >> ((1, 2, 3), frozenset({4, 5, 6}))
    hash(frozendata)
    >> -11948691520864899

Relaxing requirements (accepting functions, modules, classes and instances):
----------------------------------------------------------------------------

Functions, modules, (user-created) classes and instances are mutable in Python, and therefore
neither immutable or read-only. By default, using these will result in errors, but setting
parameter ``allow`` as one, several or all of ``functions``, ``modules`` , ``classes``
and ``instances``, these can be used in the new new data structure.

**Functions** have mutable attributes in Python, but sometimes you still want a function in a
new data structure that won't affect the parent data structure / parent function.
By setting ``allow='functions'`` or ``allow=['functions']``, the new data structure will
contain  a *copy* of the included functions and its public attributes:

.. code-block:: python

    def func(n):
        return n*2
    func.a = 'a'
    data = [func]
    frozendata = freezedata.freeze_data(data, allow='functions')
    data[0] == frozendata[0]
    >> False
    frozendata[0].a = 'b'
    print(data[0].a, frozendata[0].a)
    >> a b

**modules** will be converted to a ``namedtuple``, if you're freezing a module.
If a module is in the data structure, but it's not top level, an error will by default be raised.
If ``allow={'modules'}`` is set, non-top-level modules will be allowed and kept *unchanged*.

**classes and class instances** may be converted into ``namedtuple`` and used in the
frozen data structure by setting ``allow={'classes', 'instances}`` or only one, e.g.
``allow={'classes'}``, as needed. By converting to ``namedtuple``, information may be lost, as
attributes with leading underscores will be ignored:

.. code-block:: python

    class Test:
        a = 1
        def __init__(self, a):
            self.a = a
    test = Test(2)
    frozendata = freezedata.freeze_data([Test, test], allow={'classes', 'instances'})
    print(frozendata)
    >> (Test(a=1), Test(a=2))
    print(type(frozendata[0]), type(frozendata[1]))
    >> <class 'freezedata.freezedata.Test'> <class 'freezedata.freezedata.Test'>  # two namedtuples

"""

from .freezedata import freeze_data




import types
from collections import namedtuple


def freeze_data(obj, *, allow=frozenset()):
    """Convert 'obj' recursively to a read-only object but selectively
    allow functions, modules and other hashables, which may not be read-only.
    This means that recursively:
    - ints, floats, strings, bytes, None and bools are kept unchanged
    - lists are converted to tuples
    - dicts are converted to types.MappingProxyType
    - sets are converted to frozensets
    - modules (if allowed) will unchanged. If root object s a module, it will be converted to
      a namedtuple
    - classes and instances (if allowed) are converted to namedtuples
    - namedtuples are created anew and values recursively made read-only

    If 'obj' is a function, a ValueError is raised, as functions are not read-only
    (e.g. the function's .__globals__ will be accessible). However, functions are often useful
    to retain in the data structures, so if string 'functions' is found in 'allow',
    functions are *copied* to new functions. The user of the data structure *will* be
    allowed access to the original function's .__globals__ attributes etc., so keep this in mind.

    class objects are also not necessarily read-only and will by default raise ValueErrors.
    However, if a string 'classes' is in 'allow', the class objects will be converted to
    namedtuple instances.

    class instances (beside from built-in immutables like 9, "er", 9.7 etc.) will raise
    ValueError by default. If a string 'instances' is set in 'allow', instances will
    be converted to namedtuple instances.

    If other objects than lists, sets, instances etc mentioned above are found, a ValueError
    will be raised.

    Notice: The created read-only objects is not hashable if the original object contains dicts, as
    types.MappingProxyType is not hashable.

    """

    allowed_ = {'functions', 'modules', 'classes', 'instances'}
    if isinstance(allow, (str)):
        allow = frozenset((allow,))
    elif not isinstance(allow, (set, frozenset)):
        allow = frozenset(allow)
    if not allow <= allowed_:
        raise ValueError("'allow' must be in %r" % (allowed_))

    if isinstance(obj, types.ModuleType):
        return freeze_class_or_istance_or_module(obj, allow=allow)
    else:
        return freeze_data_inner(obj, allow=allow)


def freeze_data_inner(obj, *, allow=frozenset()):
    # check type and recursively return a new read-only object
    if isinstance(obj, (str, int, float, bytes, type(None), bool)):
        return obj
    elif isinstance(obj, tuple) and not type(obj) == tuple:  # assumed namedtuple
        return type(obj)(
            *(freeze_data_inner(i, allow=allow) for i in obj))
    elif isinstance(obj, (tuple, list)):
        return tuple(freeze_data_inner(i, allow=allow) for i in obj)
    elif isinstance(obj, (dict, types.MappingProxyType)):
        return types.MappingProxyType(
            {k: freeze_data_inner(v, allow=allow) for k, v in
             obj.items()})
    elif isinstance(obj, (set, frozenset)):
        return frozenset(
            freeze_data_inner(i, allow=allow) for i in obj)
    elif isinstance(obj, types.FunctionType):
        if 'functions' not in allow:
            raise ValueError(("Functions not allowed, %s used. "
                              "To allow functions, set allow = {'functions'}") % (
                                 obj,))
        func = copy_function(obj)
        for i in (j for j in dir(obj) if not j.startswith('__')):
            val = getattr(obj, i)
            setattr(func, i, freeze_data_inner(val, allow=allow))
        return func
    elif isinstance(obj, types.SimpleNamespace):
        return freeze_class_or_istance_or_module(obj, allow=allow)
    elif isinstance(obj, types.ModuleType):
        if 'modules' not in allow:
            raise ValueError(("Modules not allowed, %s used. "
                              "To allow functions, set allow = {'modules'}") % (
                                 obj,))
        return obj
    elif isinstance(obj, type):
        if 'classes' not in allow:
            raise ValueError(("classes not allowed, %s used. "
                              "To allow classes, set allow = {'classes'}") % (
                                 obj,))
        return freeze_class_or_istance_or_module(obj, allow=allow)
    elif isinstance(type(obj), type):
        if 'instances' not in allow:
            raise ValueError(("class instances not allowed, %s used. "
                              "To allow class instances, set allow = {'instances'}") % (
                                 obj,))
        return freeze_class_or_istance_or_module(obj, allow=allow)
    raise ValueError(obj)


def freeze_class_or_istance_or_module(obj, allow=frozenset()):
    """Convert a class, instanced or module to a collections.namedtuple.

    Note: collections.namedtuple does not allow single underscore in attributes"""
    obj_name = obj.__name__ if hasattr(obj, '__name__') else obj.__class__.__name__
    obj_name = obj_name.replace('.', '_')
    obj_tuple = namedtuple(obj_name, [i for i in dir(obj) if not i.startswith('_')])
    return obj_tuple(
        **{i: freeze_data(getattr(obj, i), allow=allow) for
           i in dir(obj) if not i.startswith('_')})


def copy_function(obj):
    """Copy function 'obj' into new new function.

    Note: functions leak data, so a function copy is only a partial copy.
    In particular, avoid resetting dunder-attributes, e.g. don't reset .__kwdefaults__ """
    import functools

    func = types.FunctionType(obj.__code__, obj.__globals__, name=obj.__name__,
                              argdefs=obj.__defaults__,
                              closure=obj.__closure__)
    func = functools.update_wrapper(func, obj)
    func.__kwdefaults__ = obj.__kwdefaults__

    return func



from setuptools import setup

description = (
'Recursively convert lists to tuples, sets to frozensets, dicts to mappingproxy etc.')


with open('README.rst') as readme:
    long_description = readme.read()

setup(name='freezedata',
      version='2.2.7',
      description=description,
      long_description=long_description,
      url='https://github.com/topper-123/freezedata',
      author='Terji Petersen',
      author_email='terji78@gmail.com',
      license='MIT License',
      packages=['freezedata'],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      test_suite='tests',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      )
